#include <stdio.h>
#include <lo/lo.h>
#include <string.h>
#include <stdlib.h>
#include <linux/input.h>
#include <errno.h>

#define rotate(a) ((a)=(a)->next)

#define SYN 0
#define DONE 0

struct net_resource {
  char *port;
  int refcount;
  lo_address addr;
};

/* node of a doubly linked circular list of specs for reporting to osc servers */
struct osc_transmitter {
  struct osc_transmitter *next;
  struct osc_transmitter *prev;
  double max;
  double value;
  char *dest;
  int type;
  int code;
  struct net_resource *net;
};

/* node of a doubly linked circular list of devices we are reading from */

char *name = NULL;
FILE *fd;


void init_trans(struct osc_transmitter *t);

struct osc_transmitter *insert_trans(struct osc_transmitter *old);

struct osc_transmitter *parse_argv(struct osc_transmitter *, int, char **);

void read_loop(struct osc_transmitter *);

int main(int argc, char *argv[]) {
  struct osc_transmitter *osc;
  osc = malloc(sizeof(struct osc_transmitter));
  osc->prev = osc; /* starting up the double circularly linked list */
  osc->next = osc;
  init_trans(osc);
  osc = parse_argv(osc, argc, argv);
  read_loop(osc);
}

void read_loop(struct osc_transmitter *osc) {
  struct osc_transmitter *start;
  struct input_event evt;
  while (!ferror(fd)) {
    fread(&evt, sizeof(struct input_event), 1, fd);
    start = osc;
    do {
      osc = osc->next; 
      if (((osc->type < 0) || (osc->type == evt.type)) &&
	  ((osc->code < 0) || (osc->code == evt.code))) {
	if (osc->net && osc->net->port != NULL) {
	  switch (osc->type) {
	  case EV_REL:
	    osc->value += evt.value;
	    if(osc->value < 0) {
	      osc->value = 0;
	    } else if (osc->value > osc->max) {
	      osc->value = osc->max;
	    }
	    break;
	  case EV_SYN:
	    osc->value = 1;
	    break;
	  case EV_KEY:
	    osc->value = evt.value;
	    if (osc->code == -1) {
	      /* this one listens to all keys from this device */
	      /* we send -key on release, key on press */
	      osc->value = evt.code * -1;
	      if (evt.value) {
		osc->value *= -1;
	      }
	    }
	    break;
	  default:
	    osc->value = evt.value;
	  }
	  lo_send(osc->net->addr, osc->dest, "f", (float)osc->value);
	} else {
	  printf("file %s:\t", name);
	  switch (evt.type) {
	  case EV_SYN: printf("SYN ");break;
	  case EV_KEY:
	    switch (evt.code) {
	    case BTN_LEFT: printf("left "); break;
	    case BTN_RIGHT: printf("right "); break;
	    case BTN_MIDDLE: printf("middle "); break;
	    case BTN_SIDE: printf("side "); break;
	    case BTN_EXTRA: printf("extra "); break;
	    case BTN_FORWARD: printf("forward "); break;
	    case BTN_BACK: printf("back "); break;
	    default: printf("a key: %d ", evt.code); break;
	    }
	    if (evt.value == 1) printf("press\t"); else printf("release\t");
	    break;
	  case EV_REL:
	    switch (evt.code) {
	    case REL_X: printf("x movement:\t%d", evt.value); break;
	    case REL_Y: printf("y movement:\t%d", evt.value); break;
	    case REL_Z: printf("z movement:\t%d", evt.value); break;
	    case REL_WHEEL: printf("wheel:\t%d", evt.value); break;
	    default: printf("relative, unbound:\t%d", evt.value);
	    } break;
	  default:
	    printf("an event:\t%d", evt.type, evt.value);
	  }
	  printf("\t-t %d, -c %d, value %d\n", evt.type, evt.code, evt.value);
	}
      }
    } while (osc != start);
  }
  exit(0);
}

void init_trans(struct osc_transmitter *t) {
  t->max   = 100.0;
  t->value = 0.0;
  t->type  = -1;
  t->code  = -1;
  t->dest  = NULL;
}

struct osc_transmitter *insert_trans(struct osc_transmitter *old) {
  struct osc_transmitter *new = malloc(sizeof(struct osc_transmitter));
  new->prev = old;
  new->next = old->next;
  new->next->prev = new;
  old->next = new;
  init_trans(new);
  return new;
}

struct osc_transmitter *parse_argv(struct osc_transmitter *osc,
				   int argc, char *argv[]){
  char switch_c = 'f';
  int run = 0;
  int i;
  for(i = 1; i < argc && !run; i++) {
    if (!strchr("bcdfkmpt", switch_c)) {
      fprintf(stderr, "unrecognized argument: %s\n", argv[i]);
      exit(1);
    }
    if (argv[i][0] == '-' && !atoi(argv[i])) {
      /* a negative number aint a switch */
      if (strlen(argv[i]) == 2) {
	  switch (argv[i][1]) {
	  case 'x':
	    osc->type = EV_REL;
	    osc->code = REL_X;
	    break;
	  case 'y':
	    osc->type = EV_REL;
	    osc->code = REL_Y;
	    break;
	  case 'w':
	    osc->type = EV_REL;
	    osc->code = REL_WHEEL;
	    break;
	  default: switch_c = argv[i][1];
	  }
      } else if (!strcmp(argv[i]+1, "dest")) {
	  switch_c = 'd';
      } else if (!strcmp(argv[i]+1, "file")) {
	switch_c = 'f';
      } else if (!strcmp(argv[i]+1, "port")) {
	switch_c = 'p';
	} else if (!strcmp(argv[i]+1, "max")) {
	switch_c = 'm';
      } else if (!strcmp(argv[i]+1, "type")) {
	switch_c = 't';
      } else if (!strcmp(argv[i]+1, "code")) {
	  switch_c = 'c';
      } else if (!strcmp(argv[i]+1, "key")) {
	switch_c = 'k';
      } else if (!strcmp(argv[i]+1, "button")) {
	switch_c = 'b';
      } else {
	fprintf(stderr, "undrecognized argument: %s\n", argv[i]);
	exit(1);
      }
    } else {
      switch (switch_c) {
      case 'd':
	if (osc->dest == NULL) {
	  osc->dest = argv[i];
	} else {
	  osc = insert_trans(osc);
	  osc->dest = argv[i];
	  osc->net = osc->prev->net;
	}
	break;
      case 'f':
	{
	  int child = 0;
	  if(name != NULL) {
	    child = fork();
	    if (child < 0) {
	      fprintf(stderr, "%s\n", strerror(errno));
	    }
	  }
	  if (child) {
	    run = 1;
	  } else {
	    struct osc_transmitter *cursor = osc->next;
	    struct osc_transmitter *cull;
	    cursor = cursor->next;
	    while(cursor != osc) {
	      cull = cursor;
	      cursor = cursor->next;
	      free(cull);
	    }
	    osc->next = osc;
	    osc->prev = osc;
	    init_trans(osc);
	    name = argv[i];
	    fd = fopen(argv[i], "rb");
	    if (fd == NULL) {
	      fprintf(stderr, "could not open %s for reading\n", argv[i]);
	      exit(1);
	    }
	  }
	}
	break;
      case 'p': {
	int found = 0;
	struct osc_transmitter *cursor = osc;
	do {
	  if (cursor->net && !strcmp(cursor->net->port, argv[i])) {
	    osc->net = cursor->net;
	    cursor->net->refcount++;
	    found = 1;
	  } else {
	    cursor = cursor->next;
	  }
	} while ((cursor != osc) && !found);
	if(!found) {
	  osc->net = malloc(sizeof(struct net_resource));
	  osc->net->refcount = 1;
	  osc->net->port = argv[i];
	  osc->net->addr = lo_address_new(NULL, osc->net->port);
	}
      }
	break;
      case 'k':
	osc->type = EV_KEY;
	if(!strcmp(argv[i], "all")) {
	  osc->code = -1;
	} else {
	  osc->code = atoi(argv[i]);
	}
	break;
      case 'b':
	osc->type = EV_KEY;
	if (!strcmp(argv[i], "left")) {
	  osc->code = BTN_LEFT;
	} else if (!strcmp(argv[i], "right")) {
	  osc->code = BTN_RIGHT;
	} else if (!strcmp(argv[i], "middle")) {
	  osc->code = BTN_MIDDLE;
	} else if (!strcmp(argv[i], "side")) {
	  osc->code = BTN_SIDE;
	} else if (!strcmp(argv[i], "extra")) {
	  osc->code = BTN_EXTRA;
	} else if (!strcmp(argv[i], "forward")) {
	  osc->code = BTN_FORWARD;
	} else if (!strcmp(argv[i], "back")) {
	  osc->code = BTN_BACK;
	} else {
	  osc->code = atoi(argv[i]);
	}
	break;
      case 'm': osc->max = atof(argv[i]); break;
      case 't': osc->type = atoi(argv[i]); break;
      case 'c': osc->code = atoi(argv[i]); break;
      }
    }
  }
  return osc;
}
